
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <vector>
#include <iostream>
#include "formula.h"
#include "probability.h"

using namespace std;

int getopt(int argc, char * const argv[], const char *optstring);
extern char *optarg;
extern int optind;
//typedef vector <double> Coordinate;

double min3(double x, double y, double z) {
  if (x<0) x = -x;
  if (y<0) y = -y;
  if (z<0) z = -z;
  if (x < y && x < z) return x;
  if (y < x && y < z) return y;
  return z;
}
  
double mydistance(vector <double> &x, vector <double> &y) {
  double d = 0;
  assert(x.size() == y.size());
  for (unsigned int i=0; i<x.size(); i++) {
    double s = min3(x[i] - y[i], x[i] +1 - y[i], y[i] +1 - x[i]);
    d += s*s;
  }
  return pow(d, 0.5);
}

/**************************************************************/
int main(int argc, char *argv[])
/**************************************************************/
{ 
  Probability p, p2; //Probabilities to choose a given variable (p) or clause (p2) 
  vector <unsigned int> clausize;   // clausize[i] = size of clause i
  Formula f;
  FILE *fout =stdout;
  char filename[100] = "";
  int opt;
  unsigned int n = 10, m = 0, kmin = 0;
  double k = 3; 
  unsigned int seed = 0;  // Random seed
  unsigned int dimension = 0;   // Dimension for dimensional models
  unsigned int expdimension = 3;// Exponent of decay in dimension model
  bool neato = false;     // Generates neato .dot file
  bool principal = false; // Remove all variables and clauses not belonging to the principal component
  bool regular = false;   // Regular model
  double beta = 0;        // beta exponent in powerlaw distribution
  double beta2 = -1;      // beta exponent in powerlaw distribution of clause sizes in powarcs model
  double alpha = 1;       // alpha base in geometric distribution
  double min = 0;
  
  while ((opt = getopt(argc, argv, "n:m:b:B:c:k:K:d:D:f:vh?s:zpr")) != -1)
    switch (opt) {
    case 'n': 
      n = atoi(optarg);
      if (n < 1) {
        fprintf(stderr, "%s: Bad variable number: %d\n",argv[0], n);
        exit(-1);
      }
      break;
    case 'm': 
      m = atoi(optarg);
      if (m < 1) {
        fprintf(stderr, "%s: Bad clause number: %d\n",argv[0], m);
        exit(-1);
      }
      break;
    case 'f': 
      strcpy(filename,optarg);
      if ((fout = fopen(filename,"w"))==NULL) {
        fprintf(stderr, "%s: Unamble to open file %s\n",argv[0], filename);
        exit(-1);
      }
      break;
    case 'k': 
      k = atof(optarg);
      if (k < 0) {
        fprintf(stderr, "%s: Bad clause size: %f\n",argv[0], k);
        exit(-1);
      }
     break;
    case 'K': 
      kmin = atoi(optarg);
      if (kmin < 0) {
        fprintf(stderr, "%s: Bad minimal clause size: %d\n",argv[0], kmin);
        exit(-1);
      }
     break;
    case 's': 
      seed = atoi(optarg);
      break;
    case 'd': 
      dimension = atoi(optarg);
      break;
    case 'D': 
      expdimension = atoi(optarg);
      break;
    case 'a': 
      alpha = atof(optarg);
      if (alpha < 1 ) {
	fprintf(stderr, "%s: Bad parameter alpha (geometric distribution): %f\n",argv[0], alpha);
	exit(-1);
      }
      break;
    case 'b': 
      beta = atof(optarg);
      if (beta < 0 ) {
	fprintf(stderr, "%s: Bad parameter beta (powerlaw distribution): %f\n",argv[0], beta);
	exit(-1);
      }
      break;
    case 'B': 
      beta2 = atof(optarg);
      if (beta2 < 0 ) {
	fprintf(stderr, "%s: Bad parameter beta (clause powerlaw distribution): %f\n",argv[0], beta2);
	exit(-1);
      }
      break;
    case 'c':
      min = atof(optarg);
      if (min < 0 || min > 1) {
	fprintf(stderr,"%s: Bad smoth factor: %f\n",argv[0], min);
	exit(-1);
      }
      break;
    case 'z':
      neato = true;
      break;
    case 'p':
      principal = true;
      break;
    case 'r':
      regular = true;
      break;
    case 'h':
    case 'v':
    case '?':
    default : printf("Usage: %s [options]\n   -n <int>\t- Number of used variables\n   -m <int>\t- Number of generated clauses\n   -k <float>\t- Clause size (or average clause size with parameter 'B')\n   -K <int>\t- Minimal clause size\n   -f <file>\t- Write to 'file' (by default write to standard output)\n   -a <float>\t- Use a geometric distribution with base 'float' for variable selection\n   -b <float>\t- Use a powerlaw distribution with exponent 'float' var variable selection\n   -B <float>\t- Use a powerlaw distribution with exponent 'float' for clause selection\n   -c <float>\t- Set smoth factor in powerlaw distribution\n   -r     \t- Use regular model\n   -d <int>\t- Use dimension model with dimension 'int'\n   -D <int>\t- Use exponent decay 'int' in dinmension model\n   -s <int>\t- Set random seed\n   -z     \t- Generates a 'dot' file for neato\n   -p     \t- Select the principal connected component\n",argv[0]);
      exit(0);
    }
  
  
  // Comon initializations

  srand(seed);
  if (m==0) m = (int)(4.25 * n);        // default m

  fprintf(stderr, "Problem size n=%d m=%d k=%f\n", n, m, k);
  
  //---------- Initialize p --------------------------------------------------------------------------------
  if (alpha > 1) {
    p.geometric(n, alpha);
    fprintf(stderr, "Using geometric distribution for variables with base %f\n", alpha);
  }
  else {
    if (beta > 0) {
      p.powerlaw(n, beta, min);
      fprintf(stderr, "Using powerlaw distribution for variables with exponent %f\n", 1.0/beta+1);
    }
    else {
      p.uniform(n);
      fprintf(stderr, "Using uniform distribution for variables\n");
    }
  }
  //---------------- In Powarc model, compute clause sizes --------------------------------------------------
  if (beta2 >=0) {
    clausize.resize(m,kmin);
    if (beta2 > 0) {
      fprintf(stderr, "Using powerlaw distribution for clause sizes with exponent %f and average size %f\n", 1.0/beta2+1, k);
      p2.powerlaw(m,beta2,min);
    }
    else {    
      fprintf(stderr, "Using binomial distribution for clause sizes with average size %f\n",k);
      p2.uniform(m);
    }
    for(unsigned int i=0; i<(k-kmin)*m; i++)
      if(++clausize[p2.choose()] >= n) {
	fprintf(stderr, "ABORT: too big clause. Decrease value of parameters 'B' or 'k'\n"); 
	exit(-1);
      }
  } 
  else {
    fprintf(stderr, "Using fixed clause sizes %d\n", (int)k);
    clausize.resize(m, (int)k);
  }
  

  //-----------------------------------------------------------------------------------------------------------
  if (dimension == 0) {
  if (!regular) {
  // ------------------------- NON REGULAR MODEL -------------------------------------------------------

  // CHOOSE m CLAUSES
    
  fprintf(stderr,"Generating clauses\n\n");
  for (unsigned int i=0; i<m; i++) {
    Clause c;
    if (i%1000 == 0)fprintf(stderr, "\e[A%3.0f%%\n", (double)i/m*100);
    
    for (unsigned int j=0; j<clausize[i]; j++) {
      Var v;
      do 
	v = (Var)p.choose();  
      while (c.addedvar(v));
      Lit l;
      c.addlit(l.mkLit(v, rand()%2==0?true:false));
    }
    f.addclause(c);
  }
  
  //-----------------------------------------------------------------------------------------------------------
  }
  else {
  // ------------------------- REGULAR MODEL -------------------------------------------------------
  fprintf(stderr, "Regular model\n");
    
  bool done = true;  // Indicates that the formula is completable
  do { 
    f.clear();
    vector <int> pk = p.floor(k * m);
    vector <int> cn(2*n);
    for (unsigned int i=0; i<n; i++) {
      cn[2*i+1] = cn[2*i] = (int)(pk[i] / 2);
      if (pk[i] % 2 == 1) 
	cn[2*i + rand()%2]++;
    }
    
    // cn[2*i-2] = number of occurrences of literal i
    // cn[2*i-1] = number of occurrences of literal -i
    
    int tv = k * m;                              // tv = number of elegible balls in the bag
    done = true;
    
    //    for (i=0; i<2*n;i++,i++)
    //      printf("%d: %d %d\n",i/2+1,cn[i],cn[i+1]);
    
    // CHOSE CLAUSES
    
    for (unsigned int i=0; i<m && done; i++) {
      int tvc = tv;              // tvc = number of elegible balls in the bag FOR THIS CLAUSE
      Clause c;
      
      for (unsigned int j=0; j<k && done; j++) {
	int r = rand() % tvc;             // randomly select one possibility
	int l = 0;                        // Compute the literal l corresponding to this ball
	while (cn[l] == 0 || c.addedvar(l/2 + 1)) l++;
	//if (l >= 2*n) {fprintf(stderr,"CRASH selected an inexistent ball\n"); exit(-1);}
	while (r >= cn[l]) {
	  r = r - cn[l];
	  l++;
	  while (cn[l]==0 || c.addedvar(l/2 + 1)) l++;
	  //if (v >= 2*n) {fprintf(stderr,"CRASH selected an inexistent ball\n");exit(-1);}
	}
	
	tv--;                                              // decrease the total number of elegible literals
	tvc = tvc - cn[(int)(l/2)*2] - cn[(int)(l/2)*2+1]; // decrease the number of elegible balls for this clause
	cn[l]--;                                           // remove the literal from the bag
	
	Lit lit;
	c.addlit(lit.mkLit(l/2+1, l%2 == 0?true:false));
	
	done = done && (tvc >= 1 || j >= k-1);
      }
      f.addclause(c);
    } 
  }
  while (!done);
  
  //-----------------------------------------------------------------------------------------------------------
  }}
  else {
  // ------------------------- DIMENSION MODEL -------------------------------------------------------

  fprintf(stderr, "Dimension model with dimension %d and probability decay %d\n", dimension, expdimension);

  //----------------- Assign coodenates to variables and clauses----------------------
  vector < vector <double> > coord(n);
  vector < vector <double> > coordclau(m);
  for (unsigned int i=0; i<n; i++) {
    coord[i].resize(dimension);
    for(unsigned int d=0; d<dimension; d++) {
      //coord[i][d] = (double)(i+0.5)/n;
      coord[i][d] = frandom();
    }
  }
  for (unsigned int i=0; i<m; i++) {
    coordclau[i].resize(dimension);
    for(unsigned int d=0; d<dimension; d++) {
      //coord[i][d] = (double)(i+0.5)/m;
      coordclau[i][d] = frandom();
    }
  }
   
  //----------------- Generation of clauses-------------------------------------------------------------------
  fprintf(stderr, "Generating clauses\n\n");
  for (unsigned int i=0; i<m; i++) {
    if (i%100 == 0)fprintf(stderr, "\e[A%3.0f%%\n", (double)i/m*100);
    
    Clause c;
    Probability q(p);
    
    for(unsigned int j=0; j<n; j++) {
      double d = max(mydistance(coordclau[i], coord[j]), 1.0/n);
      //assert(0 < d && d < 1);
      q[j] = p[j] / pow(d,expdimension);
      //fprintf(stderr, "Distance %d %d = %f p=%f\n",i, j, d,q[j]);
    }
    
    for (unsigned int j=0; j<clausize[i]; j++) {
      Var v = (Var)q.choose();  
      q[v] = 0;  // So this variable is no longer choosen
      assert(!c.addedvar(v));
      Lit l;
      c.addlit(l.mkLit(v, rand()%2==0?true:false));
    }
    f.addclause(c);
  }

  //------------------------ COMON ENDING ---------------------------------
  }

  if (principal) f.principal_component();

  if (neato) f.neato();
  else f.print();
  fclose(fout);
  exit(0);
} 
